/*
 * Copyright (c) 2003, 2004, 2006 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
#if 0
SM_RCSID("@(#)$Id: t-connctl-0.c,v 1.22 2013/10/23 22:13:36 ca Exp $")
#endif /* 0 */

/* For Solaris */
#define _STRPTIME_DONTZERO 1

#include <stdio.h>
#include "sm/ctype.h"
#include "sm/string.h"
#include "sm/time.h"
#include "sm/error.h"
#include "sm/net.h"
#include "sm/connctl.h"
#include "sm/test.h"

#define MAXCPM	100
#define MAXLINE	128

static int CollTime = 60;
static int Verbose = 0;

#ifndef CPMHSIZE
# define CPMHSIZE	256
#endif
#ifndef CT_STEPS_MAX
# define CT_STEPS_MAX	5
#endif

/* Hack... to get a "statistics" version */
#define SM_CONNCTL_PERF	1
#include "../libmta/connctl.c"

static int size = CPMHSIZE;
static int steps = CT_STEPS_MAX;

#if TEST_BHTABLE

#include "sm/bhtable.h"
#define T_BHT_SIZE	4095
#define OPEN_CONN_MAX	100
static ipv4_T open_conns[OPEN_CONN_MAX];	/* used as fifo */
static int oc_first = 0, oc_last = 0;

static void
stats(bht_P tn)
{
#if SM_BHT_PERF
	char buf[1024];

	if (Verbose > 0) {
		bht_stats(tn, buf, sizeof(buf));
		fprintf(stderr, "%s\n", buf);
	}
#endif /* SM_BHT_PERF */
}
#endif /* TEST_BHTABLE */

#define OCC_RATE 1
#define HAVE_CONNTIME 1

#include "sm/actdb-int.h"
#include "sm/occ.h"

static int
occ_rate(occ_entry_P occ_entry, time_T now)
{
	if (occ_entry->occe_max_rate > 0) {
		int cnt;
		uint u;
		ulong ticks;
		conntime_P ct;

		cnt = 0;
		ticks = now / CHTGRAN;
		ct = &occ_entry->occe_times[ticks % CHTSIZE];
		if (ct->ct_ticks != ticks) {
			ct->ct_ticks = ticks;
			ct->ct_count = 0;
		}
		++ct->ct_count;
		occ_entry->occe_rate_first = 0;
		for (u = 0; u < CHTSIZE; ++u) {
			ct = &occ_entry->occe_times[u];
			if (ct->ct_ticks <= ticks && ct->ct_ticks >= ticks - CHTSIZE) {
				cnt += ct->ct_count;
				if (0 == occ_entry->occe_rate_first)
					occ_entry->occe_rate_first = ct->ct_ticks * CHTGRAN;
			}
		}
		occ_entry->occe_rate = (cnt * 60) / (CHTSIZE * CHTGRAN);
		return occ_entry->occe_rate;
	}
	return -1;
}

/*
**  READCONN -- read connection data from file and call connctl
**
**	Parameters:
**		fp -- file from which to read data
**		rdonly -- don't call connctl
**		writeit -- write pre-processed data to stdout
**		preproc -- input is pre-processed
**
**	Returns:
**		usual sm_error code
*/

int
readconn(FILE *fp, int rdonly, int writeit, int preproc, int occrate)
{
	sm_ret_T ret;
	int r;
	char *s, *h;
	struct tm tm;
	sm_sockaddr_T addr;
	time_t t, prev;
	ulong colls, unused, lc, total, maxcoll;
	char line[MAXLINE];
#if TEST_BHTABLE
	bht_P tn;
	bht_entry_P bi;
#else
	connctx_P cctx;
	occ_entry_T occ_entry;
	time_t ntt;
	uint delay;
#endif

#if TEST_BHTABLE
	tn = bht_new(size, size * 10);
	SM_TEST_E(tn != NULL);
#else
	ret = sm_connctl_new(size, steps, &cctx);
	SM_TEST_E(ret == SM_SUCCESS);
#endif

	lc = 0;
	prev = 0;
	colls = unused = lc = total = maxcoll = 0;
	ntt = 0;
	delay = 0;
	if (occrate) {
		sm_memzero(&occ_entry, sizeof(occ_entry));
#if OCC_CHECK
		occ_entry.sm_magic = SM_OCCE_MAGIC;
#endif
		occ_entry.occe_max_rate = occrate;
	}
	while (fgets(line, sizeof(line), fp) != NULL) {
		++lc;
		if (preproc)
			t = strtol(line, &s, 10);
		else {
			bzero(&tm, sizeof(tm));
			tm.tm_year = 103;
			tm.tm_wday = 0;
			tm.tm_yday = 0;
			s = strptime(line, "%b %e %T", &tm);
			if (s == NULL) {
				fprintf(stderr, "%ld: can't read date: '%s'\n",
					lc, line);
				continue;
			}
#if HAVE_TIMEGM
			t = timegm(&tm);
#else
			t = mktime(&tm);
#endif
		}
		if (t == -1) {
			fprintf(stderr, "%ld: can't convert time: '%s' (%d)\n",
				lc, line, errno);
			continue;
		}
		if (t < prev) {
			fprintf(stderr, "%ld: time went backwards: %ld -> %ld\n",
				lc, (long) prev, (long) t);
			prev = t;
			continue;
		}
		prev = t;

		if (occrate) {
			if (ntt > 0 && ntt > t) {
				delay = ntt - t;
				ntt = 0;
			}
			if (delay > 0) {
				t += delay;
			}
		}

		while (*s != '\0' && *s != '\n' && !(isascii(*s) && isdigit(*s)))
			++s;
		h = s;
		while (*h != '\0' && *h != '\n')
			++h;
		if (*h == '\n')
			*h = '\0';
		r = sm_inet_a2sockaddr(s, NULL, &addr);
		if (sm_is_err(r)) {
			fprintf(stderr, "%ld: can't read IP addr: '%s', (%d)\n",
				lc, s, errno);
			continue;
		}
		SM_TEST_E(SM_SUCCESS == r);
		if (!rdonly) {
#if TEST_BHTABLE
			open_conns[oc_first++] = (ipv4_T) addr.s_addr;
			if (oc_first >= OPEN_CONN_MAX)
				oc_first = 0;
			if (bht_find(tn, (char *) &addr, sizeof(addr)) == NULL) {
				ret = bht_add(tn, (char *) &addr, sizeof(addr), "value", &bi);
				SM_TEST(ret == SM_SUCCESS);
				if (ret != SM_SUCCESS)
					goto done;
				SM_TEST(bi != NULL);
			}
/*
printf("first=%d, last=%d\n", oc_first, oc_last);
*/
			if ((oc_first > oc_last &&
			     oc_first - oc_last >= OPEN_CONN_MAX - 2) ||
			    (oc_first < oc_last && oc_last - oc_first <= 2))
			{
/*
printf("remove first=%d, last=%d, addr=%x\n", oc_first, oc_last, open_conns[oc_last]);
*/
				bht_rm(tn, (char *)&(open_conns[oc_last]),
					sizeof(open_conns[oc_last]),
					NULL, NULL);
				++oc_last;
				if (oc_last >= OPEN_CONN_MAX)
					oc_last = 0;
			}
#else /* TEST_BHTABLE */
			if (occrate) {
				ret = occ_rate(&occ_entry, t);
				if (occ_entry.occe_max_rate > 0 &&
					occ_entry.occe_rate > occ_entry.occe_max_rate) {

					ntt = occ_entry.occe_rate_first + 60;
					printf("rate exceeded: time=%ld rate=%d first=%ld ntt=%ld\n"
						, (long) t, occ_entry.occe_rate
						, (long) occ_entry.occe_rate_first
						, (long) ntt
						);
				} else if (Verbose > 0) {
					printf("rate: time=%ld rate=%d first=%ld ntt=%ld, delay=%u\n"
						, (long) t, occ_entry.occe_rate
						, (long) occ_entry.occe_rate_first
						, (long) ntt, delay
						);
				}
			} else {
				ret = sm_connctl(cctx, &addr, t);
			}
#endif /* TEST_BHTABLE */
		}
		if (writeit)
			printf("%ld %s\n", (long) t, s);
	}

#if TEST_BHTABLE
  done:
	stats(tn);
	bht_destroy(tn, NULL, NULL);
#else /* TEST_BHTABLE */
# if SM_CONNCTL_PERF
	ret = sm_connctl_stats(cctx, &colls, &maxcoll, &unused, &total);
# endif /* SM_CONNCTL_PERF */
	printf("total:    %7ld\n", total);
	printf("colls:    %7ld\n", colls);
	printf("maxcolls: %7ld\n", maxcoll);
	printf("unused:   %7ld\n", unused);
	sm_connctl_free(cctx);
#endif /* TEST_BHTABLE */
	return 0;

  error:
   return -1;
}

int
main(int argc, char *argv[])
{
	int i, c, rdonly, writeit, preproc, occrate;
	FILE *fp;
	const char *prg;

	prg = argv[0];
	rdonly = writeit = preproc = occrate = 0;
	while ((c = getopt(argc, argv, "c:h:o:prs:Vw")) != -1) {
		switch (c) {
		  case 'c':
			CollTime = strtol(optarg, NULL, 0);
			if (CollTime <= 1)
				return 1;
			break;
		  case 'h':
			size = strtol(optarg, NULL, 0);
			if (size <= 4)
				return 1;
			break;
		  case 'o':
			occrate = strtol(optarg, NULL, 0);
			break;
		  case 'p':
			++preproc;
			break;
		  case 'r':
			++rdonly;
			break;
		  case 's':
			steps = strtol(optarg, NULL, 0);
			if (steps <= 2 || steps >= size)
				return 1;
			break;
		  case 'V':
			++Verbose;
			break;
		  case 'w':
			++writeit;
			break;
		  default:
			return 1;
		}
	}
	sm_test_begin(argc, argv, "test connctl 0");
	argc -= optind;
	if (argc > 0) {
		for (i = 0; i < argc; i++) {
			if ((fp = fopen(argv[i + optind], "r")) != (FILE *)0) {
				readconn(fp, rdonly, writeit, preproc, occrate);
				fclose(fp);
			}
			else
				fprintf(stderr, "%s: can't read: '%s'\n", prg, argv[i]);
		}
	}
	else
		readconn(stdin, rdonly, writeit, preproc, occrate);
	return sm_test_end();
}
