/*
 * Copyright (c) 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-dkim-conf-0.c,v 1.18 2013/12/13 15:42:32 ca Exp $")

#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/stat.h"
#include "sm/fcntl.h"
#include "sm/ctype.h"
#include "sm/io.h"
#include "sm/str.h"
#include "sm/string.h"
#include "sm/mta.h"
#include "sm/misc.h"
#include "sm/base64.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#define SM_DKIMCNFDEF 1
#define SM_CONF_BYTE_DEF 1
#define SM_CONF_KBYTE_DEF 1
#include "sm/dkimcnf.h"
#include "sm/dkimcnfdef.h"
#include "sm/sysexits.h"
#include "sm/misc.h"
#include "sm/maps.h"
#include "sm/test.h"

#include "dkim.h"

#if MTA_USE_DKIM

#define BH "P7M5AsnZKX0CSpQl3w3u9XwzYQ7gDrSE5gSSQy/1W2E="

static const char *bodyhash64 = BH;
static const char *sig64 = NULL;
static const char *conffile = "dkim.conf";

static uint Verbose = 0;
static dkim_s_T dkim_s;
static dkim_cnf_T dkim_cnf;
static sm_ret_T err = 0;
static sm_ret_T conf_err = 0;
static sm_ret_T init_err = 0;
static time_T cnf_time = 0;

/*
**  DKIM_READCONF -- Read configuration file
**
**	Parameters:
**		fn -- filename of configuration file
**
**	Returns:
**		0 on success, everything else is an error
*/

static sm_ret_T
dkim_readconf(const char *fn)
{
	int err;
	uint pflags;
	sm_conf_T *smc;
	FILE *fp;

	fp = NULL;
	sm_memzero(&dkim_cnf, sizeof(dkim_cnf));
	smc = sm_conf_new(fn);
	if (smc == NULL) {
		err = errno;

		sm_io_fprintf(smioerr,
			"sev=ERROR, func=sm_readconf, file=%s, sm_conf_new=NULL, errno=%d\n",
			fn, err);
		return ENOMEM;
	}
	err = sm_conf_read_FILE(smc, fn, fp);
	if (err != 0) {
		sm_prt_conferr(fn, smc, err, smioerr);
		sm_conf_destroy(smc);
		smc = NULL;
		return err;
	}

	pflags = SM_CONF_FLAG_KEEP_DEFAULT;
	err = sm_conf_scan(smc, dkim_defs, pflags, &dkim_cnf);
	if (err != 0) {
		sm_prt_conferr(fn, smc, err, smioerr);
		return err;
	}

	return SM_SUCCESS;
}

/*
**  DKIM_TEST -- simple DKIM test
**
**	Parameters:
**		dkim_ctx -- DKIM context
**		msg -- msg
**		len -- len of msg
**		chunksize -- size of one chunk
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_test(dkim_ctx_P dkim_ctx, const uchar *msg, size_t len, ssize_t chunksize)
{
	sm_ret_T ret;
	dkim_msg_P dkim_msg;
	sm_maps_P maps_ctx;

	ret = SM_SUCCESS ;
	SM_TEST_E(msg != NULL);
	SM_TEST_E(*msg != '\0');
	SM_TEST_E(len != 0);
	SM_TEST_E(conffile != NULL);

	ret = dkim_readconf(conffile);
	if (conf_err != SM_SUCCESS && ret != SM_SUCCESS) {
		SM_TEST_E(conf_err == ret);
		goto done;
	}
	SM_TEST_E(SM_SUCCESS == ret);

	maps_ctx = NULL;
	dkim_ctx->dkc_cnf = &dkim_cnf;
	ret = dkim_conf_init(dkim_ctx, &maps_ctx);
	if (init_err != SM_SUCCESS && ret != SM_SUCCESS) {
		SM_TEST_E(init_err == ret);
		goto done;
	}
	SM_TEST_E(sm_is_success(ret));

	ret = dkim_msg_open(dkim_ctx, &dkim_msg);
	SM_TEST_E(SM_SUCCESS == ret);

	dkim_msg->dkm_s = dkim_ctx->dkc_dk_s_dflt;
	if (cnf_time != 0)
		dkim_msg->dkm_time = cnf_time;
	dkim_msg->dkm_parsefrom = dkim_ctx->dkc_dkim_map != NULL;

	if (chunksize < 0) {
		ret = dkim_msg_chunk(dkim_msg, msg, len);
		if (err != SM_SUCCESS && ret != SM_SUCCESS) {
			SM_TEST_E(err == ret);
			goto done;
		}
		else
			SM_TEST_E(SM_SUCCESS == ret);
	}
	else {
		size_t offset;
		bool last;

		for (offset = 0, last = false; !last; offset += chunksize) {
			if (chunksize + offset > len) {
				chunksize = len - offset;
				last = true;
			}
			ret = dkim_msg_chunk(dkim_msg, msg + offset, chunksize);
			if (Verbose > 2)
				sm_io_fprintf(smioerr,
					"dkim_msg=%x, offset=%u, chunksize=%ld\n",
					ret, offset, chunksize);
			if (last)
				SM_TEST_E(SM_SUCCESS == ret);
			else
				SM_TEST_E(SM_NEEDMORE == ret);
		}
	}

	if (bodyhash64 != NULL)
		SM_TEST(strstr((const char *)sm_str_getdata(dkim_msg->dkm_sighdr),
			bodyhash64) != NULL);
	if (sig64 != NULL) {
		uint u;
		sm_str_P sig;

		sig = sm_str_new(NULL, 256, 4096);	/* length??? */
		SM_TEST_E(sig != NULL);
		for (u = 0; u < sm_str_getlen(dkim_msg->dkm_sighdr); u++) {
			uchar c;

			c = sm_str_rd_elem(dkim_msg->dkm_sighdr, u);
			if (ISSPACE(c))
				continue;
			ret = sm_str_put(sig, c);
			SM_TEST_E(SM_SUCCESS == ret);
		}
		SM_TEST(strstr((const char *)sm_str_getdata(sig), sig64) != NULL);
	}
	if (Verbose > 0)
		sm_io_fprintf(smioerr, "sig: %S\n", dkim_msg->dkm_sighdr);

	ret = dkim_msg_close(dkim_msg);
	SM_TEST_E(SM_SUCCESS == ret);
  done:
	return ret;

  error:
	if (Verbose > 0)
		sm_io_fprintf(smioerr, "ret=%r\n", ret);
	return ret;
}

/*
**  USAGE -- usage message
**
**	Parameters:
**		prg -- program name
**
**	Returns:
**		none
*/

static void
usage(char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n", prg);
	sm_io_fprintf(smioerr,
		"-b text   use text as msg\n"
		"-c n      chunksize (instead of entire file at once)\n"
		"-e n      expect error n\n"
		"-E n      expect conf error n\n"
		"-I n      expect init error n\n"
		"-h hash   expected base64 bodyhash\n"
		"-k file   file contains private key for DKIM signing\n"
		"-m file   read file for msg\n"
		"-S hash   expected base64 signature\n"
		"-V        increase verbosity\n"
		);
	return;
}
#endif /* MTA_USE_DKIM */

int
main(int argc, char *argv[])
{
#if MTA_USE_DKIM
	int c, r;
	size_t msglen;
	ssize_t chunksize;
	uchar *msg;
	dkim_ctx_T dkim_ctx;
	dkim_cnf_T dkim_cnf;

	msg = NULL;
	msglen = 0;
	sm_memzero(&dkim_ctx, sizeof(dkim_ctx));
	sm_memzero(&dkim_cnf, sizeof(dkim_cnf));
	sm_memzero(&dkim_s, sizeof(dkim_s));
	dkim_ctx.dkc_cnf = &dkim_cnf;
	chunksize = -1;
	err = 0;

	while ((c = getopt(argc, argv, "Cc:h:E:e:f:I:m:pS:t:V")) != -1) {
		switch (c) {
		  case 'C':
			compoptions(smioout);
			return 0;
		  case 'c':
			chunksize = strtol(optarg, NULL, 0);
			break;
		  case 'E':
			conf_err = (sm_ret_T)strtoul(optarg, NULL, 0);
			break;
		  case 'e':
			err = (sm_ret_T)strtoul(optarg, NULL, 0);
			break;
		  case 'f':
			conffile = strdup(optarg);
			break;
		  case 'h':
			bodyhash64 = strdup(optarg);
			break;
		  case 'I':
			init_err = (sm_ret_T)strtoul(optarg, NULL, 0);
			break;
		  case 'm':
			r = readfile(optarg, &msg, &msglen);
			if (r != 0)
				return r;
			break;
		  case 'S':
			sig64 = strdup(optarg);
			break;
		  case 't':
			cnf_time = strtoull(optarg, NULL, 0);
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(argv[0]);
			return EX_USAGE;
		}
	}

	sm_test_begin(argc, argv, "test dkim conf 0");
	dkim_test(&dkim_ctx, msg, msglen, chunksize);
	return sm_test_end();
#else
	sm_io_fprintf(smioerr, "%s: SKIPPED: MTA_USE_DKIM not set\n", argv[0]);
	return 0;
#endif
}
